# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import logging
import numpy

from typing import Any, Dict, List, Set, Iterable, Iterator, Optional, Tuple, Union
from hyperopt import hp
from hyperopt.pyll import scope
from lale.util.VisitorMeta import AbstractVisitorMeta
from lale.search.PGO import FrequencyDistribution

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

PGO_input_type = Union[FrequencyDistribution, Iterable[Tuple[Any, int]], None]
class SearchSpace(metaclass=AbstractVisitorMeta):
    def __init__(self):
        pass

class SearchSpaceEnum(SearchSpace):
    pgo:Optional[FrequencyDistribution]
    vals:List[Any]

    def __init__(self, vals:Iterable[Any], pgo:PGO_input_type=None):
        super(SearchSpaceEnum, self).__init__()
        self.vals = sorted(vals, key=str)

        if pgo is None or isinstance(pgo, FrequencyDistribution):
            self.pgo = pgo
        else:
            self.pgo = FrequencyDistribution.asEnumValues(pgo, self.vals)

class SearchSpaceConstant(SearchSpaceEnum):
    def __init__(self, v, pgo:PGO_input_type=None):
        super(SearchSpaceConstant, self).__init__([v], pgo=pgo)

class SearchSpaceBool(SearchSpaceEnum):
    def __init__(self, pgo:PGO_input_type=None):
        super(SearchSpaceBool, self).__init__([True, False], pgo=pgo)

class SearchSpaceNumber(SearchSpace):
    minimum:Optional[float]
    exclusiveMinumum:bool
    maximum:Optional[float]
    exclusiveMaximum:bool
    discrete:bool
    distribution:str
    pgo:Optional[FrequencyDistribution]

    def __init__(self, 
                 minimum=None, 
                 exclusiveMinimum:bool=False, 
                 maximum=None, 
                 exclusiveMaximum:bool=False, 
                 discrete:bool=False, 
                 distribution="uniform",
                 pgo:PGO_input_type=None) -> None:
        super(SearchSpaceNumber, self).__init__()
        self.minimum = minimum
        self.exclusiveMinimum = exclusiveMinimum
        self.maximum = maximum
        self.exclusiveMaximum = exclusiveMaximum
        self.distribution = distribution
        self.discrete = discrete
        if pgo is None or isinstance(pgo, FrequencyDistribution):
            self.pgo = pgo
        else:
            if discrete:
                self.pgo = FrequencyDistribution.asIntegerValues(pgo, inclusive_min=self.getInclusiveMin(), inclusive_max=self.getInclusiveMax())
            else:
                self.pgo = FrequencyDistribution.asFloatValues(pgo, inclusive_min=self.getInclusiveMin(), inclusive_max=self.getInclusiveMax())


    def getInclusiveMax(self):
        """ Return the maximum as an inclusive maximum (exclusive maxima are adjusted accordingly)
        """
        max = self.maximum
        if self.exclusiveMaximum:
            if self.discrete:
                max = max - 1
            else:
                max = numpy.nextafter(max, float('-inf'))
        return max

    def getInclusiveMin(self):
        """ Return the maximum as an inclusive minimum (exclusive minima are adjusted accordingly)
        """
        min = self.minimum
        if self.exclusiveMinimum:
            if self.discrete:
                min = min + 1
            else:
                min = numpy.nextafter(min, float('+inf'))
        return min

class SearchSpaceArray(SearchSpace):
    def __init__(self, minimum:int=0, *, maximum:int, contents:SearchSpace, is_tuple=False) -> None:
        super(SearchSpaceArray, self).__init__()
        self.minimum = minimum
        self.maximum = maximum
        self.contents = contents
        self.is_tuple = is_tuple

class SearchSpaceList(SearchSpace):
    def __init__(self, contents:List[SearchSpace], is_tuple=False) -> None:
        super(SearchSpaceList, self).__init__()
        self.contents = contents
        self.is_tuple = is_tuple

class SearchSpaceObject(SearchSpace):
    def __init__(self, longName:str, keys:List[str], choices:Iterable[Any]) -> None:
        super(SearchSpaceObject, self).__init__()
        self.longName = longName
        self.keys = keys
        self.choices = choices



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import astunparse
import importlib
import inspect
import json
import pprint
import re

import lale.helpers
import lale.operators

def hyperparams_to_string(hps, op=None):
    if op:
        for k, v in hps.items():
            pass #TODO: use enums where possible
    def value_to_string(value):
        return pprint.pformat(value, width=10000, compact=True)
    strings = [f'{k}={value_to_string(v)}' for k, v in hps.items()]
    return ', '.join(strings)

def to_camel_case(name):
    s1 = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name)
    return s1.lower()

def indiv_op_to_string(op, name=None, module_name=None):
    assert isinstance(op, lale.operators.IndividualOp)
    if name is None:
        name = op.name()
    if module_name is None:
        import_stmt = ''
    else:
        if module_name.startswith('lale.'):
            op_name = op.name()
        else:
            op_name = op.class_name().split('.')[-1]
        if name == op_name:
            import_stmt = f'from {module_name} import {op_name}'
        else:
            import_stmt = f'from {module_name} import {op_name} as {name}'
    if hasattr(op._impl, "fit") and isinstance(op, lale.operators.TrainableIndividualOp):
        hps = hyperparams_to_string(op.hyperparams(), op)
        op_expr = f'{name}({hps})'
    else:
        op_expr = name
    if module_name is None:
        return op_expr
    else:
        return (import_stmt, op_expr)

def pipeline_to_string(pipeline, cls2name, show_imports):
    assert isinstance(pipeline, lale.operators.Pipeline)
    def shallow_copy_graph(pipeline):
        if isinstance(pipeline, lale.operators.OperatorChoice):
            return [pipeline], {pipeline:[]}, {pipeline:[]}
        steps = [*pipeline.steps()]
        preds = { step: [] for step in steps }
        succs = { step: [] for step in steps }
        for (src, dst) in pipeline.edges():
            preds[dst].append(src)
            succs[src].append(dst)
        return steps, preds, succs
    class Seq:
        def __init__(self, src, dst):
            self._src = src
            self._dst = dst
        def src(self):
            return self._src
        def dst(self):
            return self._dst
    class Par:
        def __init__(self, s0, s1):
            self._s0 = s0
            self._s1 = s1
        def s0(self):
            return self._s0
        def s1(self):
            return self._s1
    def find_seq(steps, preds, succs):
        for src in steps:
            if len(succs[src]) == 1:
                dst = succs[src][0]
                if len(preds[dst]) == 1:
                    return Seq(src, dst)
        return None
    def find_par(steps, preds, succs):
        for i0 in range(len(steps)):
            for i1 in range(i0 + 1, len(steps)):
                s0, s1 = steps[i0], steps[i1]
                preds0, preds1 = preds[s0], preds[s1]
                if len(preds0) == len(preds1) and set(preds0) == set(preds1):
                    succs0, succs1 = succs[s0], succs[s1]
                    if len(succs0)==len(succs1) and set(succs0)==set(succs1):
                        return Par(s0, s1)
        return None
    def replace_seq(old_steps, old_preds, old_succs, seq):
        new_steps, new_preds, new_succs = [], {}, {}
        for step in old_steps: #careful to keep topological order
            if step is seq.src():
                new_steps.append(seq)
                new_preds[seq] = old_preds[seq.src()]
                new_succs[seq] = old_succs[seq.dst()]
            elif step is not seq.dst():
                new_steps.append(step)
                def map_step(step):
                    if step in [seq.src(), seq.dst()]:
                        return seq
                    return step
                new_preds[step] = [map_step(pred) for pred in old_preds[step]]
                new_succs[step] = [map_step(succ) for succ in old_succs[step]]
        return new_steps, new_preds, new_succs
    def replace_par(old_steps, old_preds, old_succs, par):
        new_steps, new_preds, new_succs = [], {}, {}
        for step in old_steps: #careful to keep topological order
            if step is par.s0():
                new_steps.append(par)
                new_preds[par] = old_preds[step]
                new_succs[par] = old_succs[step]
            elif step is not par.s1():
                new_steps.append(step)
                new_preds[step] = []
                for pred in old_preds[step]:
                    if pred is par.s0():
                        new_preds[step].append(par)
                    elif pred is not par.s1():
                        new_preds[step].append(pred)
                new_succs[step] = []
                for succ in old_succs[step]:
                    if succ is par.s0():
                        new_succs[step].append(par)
                    elif succ is not par.s1():
                        new_succs[step].append(succ)
        return new_steps, new_preds, new_succs
    def introduce_structure(steps, preds, succs):
        progress = True
        while progress:
            seq = find_seq(steps, preds, succs)
            if seq:
                steps, preds, succs = replace_seq(steps, preds, succs, seq)
            par = find_par(steps, preds, succs)
            if par:
                steps, preds, succs = replace_par(steps, preds, succs, par)
            progress = seq or par
        if len(steps) == 1:
            return steps[0]
        else:
            return steps, preds, succs
    def get_module(op):
        class_name = op.class_name()
        def has_op(module_name, op_name):
            module = importlib.import_module(module_name)
            if hasattr(module, op_name):
                op = getattr(module, op_name)
                if isinstance(op, lale.operators.IndividualOp):
                    return op.class_name() == class_name
                else:
                    return hasattr(op, '__init__') and hasattr(op, 'fit') and (
                        hasattr(op, 'predict') or hasattr(op, 'transform'))
            return False
        mod_name_1 = class_name[:class_name.rfind('.')]
        mod_name_2 = mod_name_1[:mod_name_1.rfind('.')]
        if has_op(mod_name_2, op.name()):
            return mod_name_2
        elif has_op(mod_name_1, op.name()):
            return mod_name_1
        op_name = class_name[class_name.rfind('.')+1:]
        if has_op(mod_name_2, op_name):
            return mod_name_2
        assert has_op(mod_name_1, op_name)
        return mod_name_1
    class CodeGenState:
        def __init__(self):
            self.imports = []
            self.assigns = []
            self.irreducibles = []
            self.pipeline = []
            self._names = {'lale','pipeline','get_pipeline_of_applicable_type'}
            self.op2import = {}
            self.op2assign = {}
        def gensym(self, prefix):
            if prefix in self._names:
                suffix = 1
                while f'{prefix}_{suffix}' in self._names:
                    suffix += 1
                result = f'{prefix}_{suffix}'
            else:
                result = prefix
            self._names |= {result}
            return result
    gen = CodeGenState()
    def code_gen_rec(graph):
        if type(graph) is tuple:
            steps, preds, succs = graph
            dummy = gen.gensym('step')
            step2name = {}
            for step in steps:
                if isinstance(step, lale.operators.IndividualOp):
                    step2name[step] = code_gen_rec(step)
                else:
                    name = gen.gensym('step')
                    expr = code_gen_rec(step)
                    gen.irreducibles.append(f'{name} = {expr}')
                    step2name[step] = name
            make_pipeline = 'get_pipeline_of_applicable_type'
            gen.imports.append(f'from lale.operators import {make_pipeline}')
            gen.pipeline = 'pipeline = {}(\n    steps=[{}],\n    edges=[{}])' \
               .format(make_pipeline,
                       ', '.join([step2name[step] for step in steps]),
                       ', '.join([f'({step2name[src]},{step2name[tgt]})'
                                  for src in steps for tgt in succs[src]]))
            return None
        elif isinstance(graph, Seq):
            def parens(op):
                result = code_gen_rec(op)
                if isinstance(op, Par) or isinstance(op, lale.operators.OperatorChoice):
                    return f'({result})'
                return result
            return f'{parens(graph.src())} >> {parens(graph.dst())}'
        elif isinstance(graph, Par):
            def parens(op):
                result = code_gen_rec(op)
                if isinstance(op, Seq) or isinstance(op, lale.operators.OperatorChoice):
                    return f'({result})'
                return result
            return f'{parens(graph.s0())} & {parens(graph.s1())}'
        elif isinstance(graph, lale.operators.OperatorChoice):
            def parens(op):
                result = code_gen_rec(op)
                if isinstance(op, Seq) or isinstance(op, Par):
                    return f'({result})'
                return result
            printed_steps = [parens(step) for step in graph.steps()]
            return ' | '.join(printed_steps)
        elif isinstance(graph, lale.operators.IndividualOp):
            name = gen.gensym(cls2name[graph.class_name()])
            module_name = get_module(graph)
            import_stmt, op_expr = indiv_op_to_string(graph, name, module_name)
            gen.imports.append(import_stmt)
            if re.fullmatch(r'.+\(.+\)', op_expr):
                new_name = gen.gensym(to_camel_case(name))
                gen.assigns.append(f'{new_name} = {op_expr}')
                return new_name
            else:
                return name
        else:
            assert False, f'unexpected type {type} of graph {graph}'
    def code_gen_top(graph):
        expr = code_gen_rec(graph)
        if expr:
            gen.pipeline = f'pipeline = {expr}'
        code = gen.imports if show_imports else []
        code = code + gen.assigns + gen.irreducibles + [gen.pipeline]
        result = '\n'.join(code)
        return result
    steps, preds, succs = shallow_copy_graph(pipeline)
    graph = introduce_structure(steps, preds, succs)
    return code_gen_top(graph)

def schema_to_string(schema):
    s1 = json.dumps(schema)
    s2 = ast.parse(s1)
    s3 = astunparse.unparse(s2)
    s4 = s3.strip()
    s5 = s4
    while True:
        s6 = re.sub(r',\n\s+([\]}][^ ])', r'\1', s5)
        if s5 == s6:
            break
        s5 = s6
    s7 = re.sub(r'{\s+}', r'{}', s6)
    return s7

def to_string(arg, show_imports=True, call_depth=2):
    def get_cls2name():
        frame = inspect.stack()[call_depth][0]
        result = {}
        all_items = [*frame.f_locals.items(), *frame.f_globals.items()]
        for nm, op in all_items:
            if isinstance(op, lale.operators.IndividualOp) and nm[0].isupper():
                cls = op.class_name()
                if cls not in result:
                    result[cls] = nm
        return result
    if lale.helpers.is_schema(arg):
        return schema_to_string(arg)
    elif isinstance(arg, lale.operators.IndividualOp):
        return indiv_op_to_string(arg)
    elif isinstance(arg, lale.operators.Pipeline):
        return pipeline_to_string(arg, get_cls2name(), show_imports)
    else:
        raise ValueError(f'Unexpected argument type {type(arg)} for {arg}')

def ipython_display(arg, show_imports=True):
    import IPython.display
    pretty_printed = to_string(arg, show_imports, call_depth=3)
    markdown = IPython.display.Markdown(f'```python\n{pretty_printed}\n```')
    IPython.display.display(markdown)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import jsonschema
import warnings
from lale.lib.lale import ConcatFeatures
from lale.lib.lale import NoOp
from lale.lib.sklearn import KNeighborsClassifier
from lale.lib.sklearn import LinearSVC
from lale.lib.sklearn import LogisticRegression
from lale.lib.sklearn import MinMaxScaler
from lale.lib.sklearn import MLPClassifier
from lale.lib.sklearn import Nystroem
from lale.lib.sklearn import OneHotEncoder
from lale.lib.sklearn import PCA
from lale.lib.sklearn import TfidfVectorizer
from lale.lib.sklearn import MultinomialNB
from lale.lib.sklearn import SimpleImputer
from lale.lib.sklearn import SVC
from lale.lib.xgboost import XGBClassifier
from lale.lib.sklearn import PassiveAggressiveClassifier
from lale.lib.sklearn import StandardScaler
from lale.lib.sklearn import FeatureAgglomeration
from lale.search.GridSearchCV import LaleGridSearchCV, get_grid_search_parameter_grids

class TestClassification(unittest.TestCase):

    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split
        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test =  train_test_split(X, y)    

def create_function_test_classifier(clf_name):
    def test_classifier(self):
        X_train, y_train = self.X_train, self.y_train
        X_test, y_test = self.X_test, self.y_test
        import importlib
        module_name = ".".join(clf_name.split('.')[0:-1])
        class_name = clf_name.split('.')[-1]
        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        clf = class_()

        #test_schemas_are_schemas
        from lale.helpers import validate_is_schema
        validate_is_schema(clf.input_schema_fit())
        validate_is_schema(clf.input_schema_predict())
        validate_is_schema(clf.output_schema())
        validate_is_schema(clf.hyperparam_schema())

        #test_init_fit_predict
        trained = clf.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)

        #test_with_hyperopt
        from lale.lib.lale import HyperoptClassifier
        hyperopt = HyperoptClassifier(model = clf, max_evals=1)
        trained = hyperopt.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)

        #test_cross_validation
        from lale.helpers import cross_val_score
        cv_results = cross_val_score(clf, X_train, y_train, cv = 2)
        self.assertEqual(len(cv_results), 2)

        #test_with_gridsearchcv_auto_wrapped
        from sklearn.metrics import accuracy_score, make_scorer
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            from lale.lib.sklearn.gradient_boosting_classifier import GradientBoostingClassifierImpl
            from lale.lib.sklearn.mlp_classifier import MLPClassifierImpl
            if isinstance(clf._impl, GradientBoostingClassifierImpl):
                #because exponential loss does not work with iris dataset as it is not binary classification
                import lale.schemas as schemas
                clf = clf.customize_schema(loss=schemas.Enum(default='deviance', values=['deviance']))
            if not isinstance(clf._impl, MLPClassifierImpl):
                #mlp fails due to issue #164.
                grid_search = LaleGridSearchCV(clf, lale_num_samples=1, lale_num_grids=1, cv=2,
                                scoring=make_scorer(accuracy_score))
                grid_search.fit(X_train, y_train)

        #test_predict_on_trainable
        trained = clf.fit(X_train, y_train)
        clf.predict(X_train)

        #test_to_json
        clf.to_json()

        #test_in_a_pipeline
        pipeline = NoOp() >> clf
        trained = pipeline.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)


    test_classifier.__name__ = 'test_{0}'.format(clf.split('.')[-1])
    return test_classifier

classifiers = ['lale.lib.sklearn.RandomForestClassifier',
               'lale.lib.sklearn.DecisionTreeClassifier',
               'lale.lib.sklearn.ExtraTreesClassifier',
               'lale.lib.sklearn.GradientBoostingClassifier',
               'lale.lib.sklearn.GaussianNB',
               'lale.lib.sklearn.QuadraticDiscriminantAnalysis',
               'lale.lib.lightgbm.LGBMClassifier',
               'lale.lib.xgboost.XGBClassifier',
               'lale.lib.sklearn.KNeighborsClassifier',
               'lale.lib.sklearn.LinearSVC',
               'lale.lib.sklearn.LogisticRegression',
               'lale.lib.sklearn.MLPClassifier',
               'lale.lib.sklearn.SVC',
               'lale.lib.sklearn.PassiveAggressiveClassifier',
               'lale.lib.sklearn.MultinomialNB']
for clf in classifiers:
    setattr(
        TestClassification,
        'test_{0}'.format(clf.split('.')[-1]),
        create_function_test_classifier(clf)
    )

class TestRegression(unittest.TestCase):

    def setUp(self):
        from sklearn.datasets import make_regression
        from sklearn.model_selection import train_test_split
        X, y = make_regression(n_features=4, n_informative=2,
                               random_state=0, shuffle=False)
        self.X_train, self.X_test, self.y_train, self.y_test =  train_test_split(X, y)    

def create_function_test_regressor(clf_name):
    def test_regressor(self):
        X_train, y_train = self.X_train, self.y_train
        X_test, y_test = self.X_test, self.y_test
        import importlib
        module_name = ".".join(clf_name.split('.')[0:-1])
        class_name = clf_name.split('.')[-1]
        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        regr = class_()

        #test_schemas_are_schemas
        from lale.helpers import validate_is_schema
        validate_is_schema(regr.input_schema_fit())
        validate_is_schema(regr.input_schema_predict())
        validate_is_schema(regr.output_schema())
        validate_is_schema(regr.hyperparam_schema())

        #test_init_fit_predict
        trained = regr.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)

        #test_predict_on_trainable
        trained = regr.fit(X_train, y_train)
        regr.predict(X_train)

        #test_to_json
        regr.to_json()

        #test_in_a_pipeline
        pipeline = NoOp() >> regr
        trained = pipeline.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)

        #test_with_hyperopt
        from lale.lib.sklearn.ridge import RidgeImpl
        if not isinstance(regr._impl, RidgeImpl):
            from lale.lib.lale import HyperoptRegressor
            hyperopt = HyperoptRegressor(model = pipeline, max_evals=1)
            trained = hyperopt.fit(self.X_train, self.y_train)
            predictions = trained.predict(self.X_test)

    test_regressor.__name__ = 'test_{0}'.format(clf_name.split('.')[-1])
    return test_regressor

regressors = ['lale.lib.sklearn.RandomForestRegressor',
              'lale.lib.sklearn.DecisionTreeRegressor',
              'lale.lib.sklearn.ExtraTreesRegressor',
              'lale.lib.sklearn.GradientBoostingRegressor',
              'lale.lib.sklearn.LinearRegression',
              'lale.lib.sklearn.Ridge',
              'lale.lib.lightgbm.LGBMRegressor',
              'lale.lib.xgboost.XGBRegressor']
for clf in regressors:
    setattr(
        TestRegression,
        'test_{0}'.format(clf.split('.')[-1]),
        create_function_test_regressor(clf)
    )

class TestFeaturePreprocessing(unittest.TestCase):

    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split
        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test =  train_test_split(X, y)    

def create_function_test_feature_preprocessor(fproc_name):
    def test_feature_preprocessor(self):
        X_train, y_train = self.X_train, self.y_train
        X_test, y_test = self.X_test, self.y_test
        import importlib
        module_name = ".".join(fproc_name.split('.')[0:-1])
        class_name = fproc_name.split('.')[-1]
        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        fproc = class_()

        from lale.lib.sklearn.one_hot_encoder import OneHotEncoderImpl
        if isinstance(fproc._impl, OneHotEncoderImpl):
            #fproc = OneHotEncoder(handle_unknown = 'ignore')
            #remove the hack when this is fixed
            fproc = PCA()
        #test_schemas_are_schemas
        from lale.helpers import validate_is_schema
        validate_is_schema(fproc.input_schema_fit())
        validate_is_schema(fproc.input_schema_transform())
        validate_is_schema(fproc.output_schema())
        validate_is_schema(fproc.hyperparam_schema())

        #test_init_fit_transform
        trained = fproc.fit(self.X_train, self.y_train)
        predictions = trained.transform(self.X_test)

        #test_predict_on_trainable
        trained = fproc.fit(X_train, y_train)
        fproc.transform(X_train)

        #test_to_json
        fproc.to_json()

        #test_in_a_pipeline
        #This test assumes that the output of feature processing is compatible with LogisticRegression
        from lale.lib.sklearn import LogisticRegression
        pipeline = fproc >> LogisticRegression()
        trained = pipeline.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)

        #Tune the pipeline with LR using HyperoptClassifier
        from lale.lib.lale import HyperoptClassifier
        hyperopt = HyperoptClassifier(model = pipeline, max_evals=1)
        trained = hyperopt.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)

    test_feature_preprocessor.__name__ = 'test_{0}'.format(fproc_name.split('.')[-1])
    return test_feature_preprocessor

feature_preprocessors = ['lale.lib.sklearn.PolynomialFeatures',
                         'lale.lib.sklearn.PCA',
                         'lale.lib.sklearn.Nystroem',
                         'lale.lib.sklearn.Normalizer',
                         'lale.lib.sklearn.MinMaxScaler',
                         'lale.lib.sklearn.OneHotEncoder',
                         'lale.lib.sklearn.SimpleImputer',
                         'lale.lib.sklearn.StandardScaler',
                         'lale.lib.sklearn.FeatureAgglomeration',
                         'lale.lib.sklearn.RobustScaler',
                         ]
for fproc in feature_preprocessors:
    setattr(
        TestFeaturePreprocessing,
        'test_{0}'.format(fproc.split('.')[-1]),
        create_function_test_feature_preprocessor(fproc)
    )

class TestConcatFeatures(unittest.TestCase):
    def test_hyperparam_defaults(self):
        cf = ConcatFeatures()
    def test_init_fit_predict(self):
        trainable_cf = ConcatFeatures()
        A = [ [11, 12, 13],
              [21, 22, 23],
              [31, 32, 33] ]
        B = [ [14, 15],
              [24, 25],
              [34, 35] ]

        trained_cf = trainable_cf.fit(X = [A, B])
        transformed = trained_cf.transform([A, B])
        expected = [ [11, 12, 13, 14, 15],
                     [21, 22, 23, 24, 25],
                     [31, 32, 33, 34, 35] ]
        for i_sample in range(len(transformed)):
            for i_feature in range(len(transformed[i_sample])):
                self.assertEqual(transformed[i_sample][i_feature],
                                 expected[i_sample][i_feature])
    def test_comparison_with_scikit(self):
        import warnings
        warnings.filterwarnings("ignore")
        from lale.lib.sklearn import PCA
        import sklearn.datasets
        from lale.helpers import cross_val_score
        pca = PCA(n_components=3, random_state=42, svd_solver='arpack')
        nys = Nystroem(n_components=10, random_state=42)
        concat = ConcatFeatures()
        lr = LogisticRegression(random_state=42, C=0.1)
        trainable = (pca & nys) >> concat >> lr
        digits = sklearn.datasets.load_digits()
        X, y = sklearn.utils.shuffle(digits.data, digits.target, random_state=42)

        cv_results = cross_val_score(trainable, X, y)
        cv_results = ['{0:.1%}'.format(score) for score in cv_results]

        from sklearn.pipeline import make_pipeline, FeatureUnion
        from sklearn.decomposition import PCA as SklearnPCA
        from sklearn.kernel_approximation import Nystroem as SklearnNystroem
        from sklearn.linear_model import LogisticRegression as SklearnLR
        from sklearn.model_selection import cross_val_score
        union = FeatureUnion([("pca", SklearnPCA(n_components=3, random_state=42, svd_solver='arpack')),
                            ("nys", SklearnNystroem(n_components=10, random_state=42))])
        lr = SklearnLR(random_state=42, C=0.1)
        pipeline = make_pipeline(union, lr)

        scikit_cv_results = cross_val_score(pipeline, X, y, cv = 5)
        scikit_cv_results = ['{0:.1%}'.format(score) for score in scikit_cv_results]
        self.assertEqual(cv_results, scikit_cv_results)
        warnings.resetwarnings()
    def test_with_pandas(self):
        from lale.datasets import load_iris_df
        import warnings
        warnings.filterwarnings("ignore")
        pca = PCA(n_components=3)
        nys = Nystroem(n_components=10)
        concat = ConcatFeatures()
        lr = LogisticRegression(random_state=42, C=0.1)
        trainable = (pca & nys) >> concat >> lr

        (X_train, y_train), (X_test, y_test) = load_iris_df()
        trained = trainable.fit(X_train, y_train)
        predicted = trained.predict(X_test)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import logging
import numpy

from typing import Any, Dict, List, Set, Iterable, Iterator, Optional, Tuple, Union
from lale.schema_simplifier import findRelevantFields, narrowToGivenRelevantFields, simplify, filterForOptimizer

from lale.schema_utils import Schema, getMinimum, getMaximum, STrue, SFalse, is_false_schema, is_true_schema
from lale.search.search_space import *
from lale.search.HP import search_space_to_str_for_comparison
from lale.search.PGO import PGO, FrequencyDistribution, Freqs

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

class FreqsWrapper(object):
    base:Optional[Dict[str,Freqs]]

    def __init__(self, base:Optional[Dict[str,Freqs]]):
        self.base = base

def pgo_lookup(pgo:Optional[PGO], name:str)->Optional[FreqsWrapper]:
    if pgo is None:
        return None
    else:
        freqs:Optional[Dict[str,Freqs]] = None
        if pgo is not None:
            freqs = pgo.get(name, None)
        return FreqsWrapper(freqs)

pgo_part = Union[FreqsWrapper, Freqs, None]

def freqs_wrapper_lookup(part:pgo_part, k:str)->pgo_part:
    if part is None:
        return None
    elif isinstance(part, FreqsWrapper):
        f = part.base
        if f is not None and k in f:
            return f[k]
        else:
            return None
    else:
        return None

def asFreqs(part:pgo_part)->Optional[Iterable[Tuple[Any, int]]]:
    if part is None:
        return None
    elif isinstance(part, FreqsWrapper):
        return None
    else:
        return part.items()

def schemaObjToSearchSpaceHelper(
    longName:str, 
    path:str, 
    schema:Schema, 
    relevantFields:Optional[Set[str]],
    pgo_freqs:pgo_part=None)->Dict[str,SearchSpace]:
        if 'properties' not in schema:
            return {}
        props = schema['properties']
        hyp:Dict[str, SearchSpace] = {}
        for p,s in props.items():
            if relevantFields is None or p in relevantFields:
                # TODO: This does not handle nested relevant fields correctly
                # We would need to specify what is correct in that case
                sub_freqs = freqs_wrapper_lookup(pgo_freqs, p)
                sub_sch = schemaToSearchSpaceHelper_(longName, path + "_" + p, s, relevantFields, pgo_freqs=sub_freqs)
                if sub_sch is None:
                    # if it is a required field, this entire thing should be None
                    hyp[p] = SearchSpaceConstant(None)
                else:
                    hyp[p] = sub_sch
            else:
                logger.debug(f"schemaToSearchSpace: skipping not relevant field {p}")
        return hyp

def schemaToSearchSpaceHelper_( longName, 
                                path:str, 
                                schema:Schema, 
                                relevantFields:Optional[Set[str]],
                                pgo_freqs:pgo_part=None)->Optional[SearchSpace]:
    # TODO: handle degenerate cases
    # right now, this handles only a very fixed form

    if is_false_schema(schema):
        return None

    if 'enum' in schema:
        vals = schema['enum']
        return SearchSpaceEnum(vals, pgo=asFreqs(pgo_freqs))

    if 'type' in schema:
        typ = schema['type']
        if typ == "boolean":
            return SearchSpaceBool(pgo=asFreqs(pgo_freqs))
        elif typ == "number" or typ == "integer":
            exclusive_minimum = False
            minimum=schema.get('minimumForOptimizer', None)
            if minimum is not None:
                exclusive_minimum = schema.get('exclusiveMinimumForOptimizer', False)
            else:
                minimum=schema.get('minimum', None)
                if minimum is not None:
                    exclusive_minimum = schema.get('exclusiveMinimum', False)

            exclusive_maximum = False
            maximum=schema.get('maximumForOptimizer', None)
            if maximum is not None:
                exclusive_maximum = schema.get('exclusiveMaximumForOptimizer', False)
            else:
                maximum=schema.get('maximum', None)
                if maximum is not None:
                    exclusive_maximum = schema.get('exclusiveMaximum', False)

            distribution = schema.get('distribution', None)

            typeForOptimizer = schema.get('typeForOptimizer', None)
            if typeForOptimizer is None:
                typeForOptimizer = typ

            if typeForOptimizer == "number":
                discrete = False
            elif typeForOptimizer == "integer":
                discrete = True
            else:
                raise NotImplementedError()
            
            pgo:Freqs

            return SearchSpaceNumber(minimum=minimum, 
                            exclusiveMinimum=exclusive_minimum, 
                            maximum=maximum, 
                            exclusiveMaximum=exclusive_maximum, 
                            discrete=discrete, 
                            distribution=distribution,
                            pgo=asFreqs(pgo_freqs))
        elif typ == "array" or typ =="tuple":
            typeForOptimizer = schema.get('typeForOptimizer', None)
            if typeForOptimizer is None:
                typeForOptimizer = typ

            is_tuple:bool = typeForOptimizer == "tuple"

            items_schema = schema.get('itemsForOptimizer', None)
            if items_schema is None:
                items_schema = schema.get('items', None)
                if items_schema is None:
                    raise ValueError(f"an array type was found without a provided schema for the items in the schema {schema}.  Please provide a schema for the items (consider using itemsForOptimizer)")

            if isinstance(items_schema, list):
                contents = []
                for i,sub_schema in enumerate(items_schema):
                    sub = schemaToSearchSpaceHelper_(longName, path + "_" + str(i), sub_schema, relevantFields)
                    if sub is None:
                        return None
                    else:
                        contents.append(sub)
                return SearchSpaceList(contents=contents, is_tuple=is_tuple)

            min_items = schema.get('minItemsForOptimizer', None)
            if min_items is None:
                min_items = schema.get('minItems', None)
                if min_items is None:
                    min_items = 0
            max_items = schema.get('maxItemsForOptimizer', None)
            if max_items is None:
                max_items = schema.get('maxItems', None)
                if max_items is None:
                    raise ValueError(f"an array type was found without a provided maximum number of items in the schema {schema}.  Please provide a maximum (consider using maxItemsForOptimizer)")

            sub_opt = schemaToSearchSpaceHelper_(longName, path + "-", items_schema, relevantFields)
            is_tuple = typeForOptimizer == "tuple"
            if sub_opt is None:
                if min_items <= 0 and max_items > 0:
                    return SearchSpaceConstant([])
                else:
                    return None
            else:
                return SearchSpaceArray(minimum=min_items, maximum=max_items, contents=sub_opt, is_tuple=is_tuple)

        elif typ == "object":
            if 'properties' not in schema:
                return SearchSpaceObject(longName, [], [])
            o = schemaObjToSearchSpaceHelper(longName, path, schema, relevantFields, pgo_freqs=pgo_freqs)
            all_keys = list(o.keys())
            all_keys.sort()
            o_choice = tuple([o.get(k, None) for k in all_keys])
            return SearchSpaceObject(longName, all_keys, [o_choice])
        elif typ == "string":
            pass
        else:
            raise ValueError(f"An unknown type ({typ}) was found in the schema {schema}")


    if 'anyOf' in schema:
        objs = []
        for s_obj in schema['anyOf']:
            if 'type' in s_obj and s_obj['type'] == "object":
                o = schemaObjToSearchSpaceHelper(longName, path, s_obj, relevantFields, pgo_freqs=pgo_freqs)
                if o: objs.append(o)
        if objs:
            # First, gather a list of all the properties
            keys_list = [set(o.keys()) for o in objs]
            # make sure the iterator is deterministic
            all_keys = list(set.union(*keys_list))
            # and we might as well make it sorted
            all_keys.sort()
            def as_str(k, c):
                if c is None:
                    return "None"
                else:
                    return search_space_to_str_for_comparison(c, path + "_" + k)
            anys:Dict[str,Any] = {}
            for o in objs:
                o_choice = tuple([o.get(k, None) for k in all_keys])
                k = str([as_str(all_keys[idx], c) for idx, c in enumerate(o_choice)])
                if k in anys:
                    logger.info(f"Ignoring Duplicate SearchSpace entry {k}")
                anys[k] = o_choice
            return SearchSpaceObject(longName, all_keys, anys.values())
        else:
            return SearchSpaceObject(longName, [], [])
    
    if 'allOf' in schema:
        # if all but one are negated constraints, we will just ignore them
        pos_sub_schema:List[Schema] = []
        for sub_schema in schema['allOf']:
            if 'not' not in sub_schema:
                pos_sub_schema.append(sub_schema)

        if len(pos_sub_schema) > 1:
            raise ValueError(f"schemaToSearchSpaceHelper does not yet know how to compile the given schema {schema} for {longName}, because it is an allOf with more than one non-negated schemas ({pos_sub_schema})")
        if len(pos_sub_schema) == 0:
            raise ValueError(f"schemaToSearchSpaceHelper does not yet know how to compile the given schema {schema} for {longName}, because it is an allOf with only negated schemas")
        
        logger.debug(f"schemaToSearchSpaceHelper: ignoring negated schemas in the conjunction {schema} for {longName}")
        return schemaToSearchSpaceHelper_(longName, 
                                path,
                                pos_sub_schema[0], 
                                relevantFields,
                                pgo_freqs=pgo_freqs)
    # TODO: handle degenerate cases
    raise ValueError(f"schemaToSearchSpaceHelper does not yet know how to compile the given schema {schema} for {longName}")

def schemaToSearchSpaceHelper(longName, 
                              schema:Schema, 
                              relevantFields:Optional[Set[str]],
                              pgo_freqs:pgo_part=None)->Optional[SearchSpace]:
    if not is_false_schema(schema) and not schema:
        return None
    else:
        return schemaToSearchSpaceHelper_(longName, longName, schema, relevantFields, pgo_freqs=pgo_freqs)

def schemaToSimplifiedAndSearchSpace(
    longName:str, 
    name:str, 
    schema:Schema,
    pgo:Optional[PGO]=None)->Tuple[Schema, Optional[SearchSpace]]:
    relevantFields = findRelevantFields(schema)
    if relevantFields:
        schema = narrowToGivenRelevantFields(schema, relevantFields)
    simplified_schema = simplify(schema, True)
#    from . import helpers
#    helpers.print_yaml('SIMPLIFIED_' + longName, simplified_schema)

    filtered_schema = filterForOptimizer(simplified_schema)
#    helpers.print_yaml('FILTERED_' + longName, filtered_schema)

    return (filtered_schema, 
            schemaToSearchSpaceHelper(
                        longName, 
                        filtered_schema, 
                        relevantFields, 
                        pgo_freqs=pgo_lookup(pgo, name)))

def schemaToSearchSpace(longName:str, 
                        name:str, 
                        schema:Schema, 
                        pgo:Optional[PGO]=None)->Optional[SearchSpace]:
    (s, h) = schemaToSimplifiedAndSearchSpace(longName, name, schema, pgo=pgo)
    return h



import unittest

class TestTextEncoders(unittest.TestCase):

    def setUp(self):

        self.X_train = ['Boston locates in the East Coast',
                        'Boston Celtics is part of the East conference of NBA',
                        'Cambridge is part of the Greater Boston Area',
                        'Manhattan is located in the lower part of NYC',
                        'People worked at New York city usually lives in New Jersey Area'
                        'The financial center in the world is New York']

        self.y_train = [0, 0, 0, 1, 1, 1]

def create_function_test_encoder(encoder_name):
    def test_encoder(self):
        import importlib
        module_name = ".".join(encoder_name.split('.')[0:-1])
        class_name = encoder_name.split('.')[-1]
        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        encoder = class_()

        #test_schemas_are_schemas
        from lale.helpers import validate_is_schema
        validate_is_schema(encoder.input_schema_fit())
        validate_is_schema(encoder.input_schema_predict())
        validate_is_schema(encoder.output_schema())
        validate_is_schema(encoder.hyperparam_schema())

        #test_init_fit_transform
        trained = encoder.fit(self.X_train, self.y_train)
        transformed = trained.transform(self.X_train)

    test_encoder.__name__ = 'test_{0}'.format(encoder_name.split('.')[-1])
    return test_encoder

encoders = ['lale.lib.tensorflow.USEPretrainedEncoder']
for encoder in encoders:
    setattr(
        TestTextEncoders,
        'test_{0}'.format(encoder.split('.')[-1]),
        create_function_test_encoder(encoder)
    )




from sklearn.kernel_approximation import AdditiveChi2Sampler as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class AdditiveChi2SamplerImpl():

    def __init__(self, sample_steps=2, sample_interval=None):
        self._hyperparams = {
            'sample_steps': sample_steps,
            'sample_interval': sample_interval}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def transform(self, X):
        return self._sklearn_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for AdditiveChi2Sampler    Approximate feature map for additive chi2 kernel.',
    'allOf': [{
        'type': 'object',
        'required': ['sample_steps', 'sample_interval'],
        'relevantToOptimizer': ['sample_steps'],
        'additionalProperties': False,
        'properties': {
            'sample_steps': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 3,
                'distribution': 'uniform',
                'default': 2,
                'description': 'Gives the number of (complex) sampling points.'},
            'sample_interval': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Sampling interval. Must be specified when sample_steps not in {1,2,3}.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Set the parameters',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data, where n_samples in the number of samples'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply approximate feature map to X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Whether the return value is an array of sparse matrix depends on',
    'XXX TODO XXX': '{array, sparse matrix},                shape = (n_samples, n_features * (2*sample_steps + 1))',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
AdditiveChi2Sampler = lale.operators.make_operator(AdditiveChi2SamplerImpl, _combined_schemas)





from sklearn.tree.tree import DecisionTreeRegressor as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class DecisionTreeRegressorImpl():

    def __init__(self, criterion='mse', splitter='best', max_depth=None, min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features=None, random_state=None, max_leaf_nodes=None, min_impurity_decrease=0.0, min_impurity_split=None, presort=False):
        self._hyperparams = {
            'criterion': criterion,
            'splitter': splitter,
            'max_depth': max_depth,
            'min_samples_split': min_samples_split,
            'min_samples_leaf': min_samples_leaf,
            'min_weight_fraction_leaf': min_weight_fraction_leaf,
            'max_features': max_features,
            'random_state': random_state,
            'max_leaf_nodes': max_leaf_nodes,
            'min_impurity_decrease': min_impurity_decrease,
            'min_impurity_split': min_impurity_split,
            'presort': presort}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def predict(self, X):
        return self._sklearn_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for DecisionTreeRegressor    A decision tree regressor.',
    'allOf': [{
        'type': 'object',
        'required': ['criterion', 'splitter', 'max_depth', 'min_samples_split', 'min_samples_leaf', 'min_weight_fraction_leaf', 'max_features', 'random_state', 'max_leaf_nodes', 'min_impurity_decrease', 'min_impurity_split', 'presort'],
        'relevantToOptimizer': ['criterion', 'splitter', 'max_depth', 'min_samples_split', 'min_samples_leaf', 'max_features'],
        'additionalProperties': False,
        'properties': {
            'criterion': {
                'enum': ['friedman_mse', 'mse'],
                'default': 'mse',
                'description': 'The function to measure the quality of a split. Supported criteria'},
            'splitter': {
                'enum': ['random', 'best'],
                'default': 'best',
                'description': 'The strategy used to choose the split at each node. Supported'},
            'max_depth': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 3,
                    'maximumForOptimizer': 5,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The maximum depth of the tree. If None, then nodes are expanded until'},
            'min_samples_split': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 5,
                    'distribution': 'uniform'}, {
                    'type': 'number',
                    'forOptimizer': False}],
                'default': 2,
                'description': 'The minimum number of samples required to split an internal node:'},
            'min_samples_leaf': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 1,
                    'maximumForOptimizer': 5,
                    'distribution': 'uniform'}, {
                    'type': 'number',
                    'forOptimizer': False}],
                'default': 1,
                'description': 'The minimum number of samples required to be at a leaf node.'},
            'min_weight_fraction_leaf': {
                'type': 'number',
                'default': 0.0,
                'description': 'The minimum weighted fraction of the sum total of weights (of all'},
            'max_features': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'}, {
                    'type': 'string',
                    'forOptimizer': False}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of features to consider when looking for the best split:'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator;'},
            'max_leaf_nodes': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Grow a tree with ``max_leaf_nodes`` in best-first fashion.'},
            'min_impurity_decrease': {
                'type': 'number',
                'default': 0.0,
                'description': 'A node will be split if this split induces a decrease of the impurity'},
            'min_impurity_split': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Threshold for early stopping in tree growth. A node will split'},
            'presort': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to presort the data to speed up the finding of best splits in'},
        }}, {
        'XXX TODO XXX': 'Parameter: min_samples_leaf > only be considered if it leaves at least min_samples_leaf training samples in each of the left and right branches'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Build a decision tree regressor from the training set (X, y).',
    'type': 'object',
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix, shape = [n_samples, n_features]'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'The training input samples. Internally, it will be converted to'},
        'y': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'The target values (real numbers). Use ``dtype=np.float64`` and'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'description': 'Sample weights. If None, then samples are equally weighted. Splits'},
        'check_input': {
            'type': 'boolean',
            'default': True,
            'description': 'Allow to bypass several input checking.'},
        'X_idx_sorted': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}, {
                'enum': [None]}],
            'default': None,
            'description': 'The indexes of the sorted training input samples. If many tree'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class or regression value for X.',
    'type': 'object',
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix of shape = [n_samples, n_features]'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'The input samples. Internally, it will be converted to'},
        'check_input': {
            'type': 'boolean',
            'default': True,
            'description': 'Allow to bypass several input checking.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The predicted classes, or the predict values.',
    'anyOf': [{
        'type': 'array',
        'items': {
            'type': 'number'},
    }, {
        'type': 'array',
        'items': {
            'type': 'array',
            'items': {
                'type': 'number'},
        }}],
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
DecisionTreeRegressor = lale.operators.make_operator(DecisionTreeRegressorImpl, _combined_schemas)





from sklearn.svm.classes import LinearSVR as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class LinearSVRImpl():

    def __init__(self, epsilon=0.0, tol=0.0001, C=1.0, loss='epsilon_insensitive', fit_intercept=True, intercept_scaling=1.0, dual=True, verbose=0, random_state=None, max_iter=1000):
        self._hyperparams = {
            'epsilon': epsilon,
            'tol': tol,
            'C': C,
            'loss': loss,
            'fit_intercept': fit_intercept,
            'intercept_scaling': intercept_scaling,
            'dual': dual,
            'verbose': verbose,
            'random_state': random_state,
            'max_iter': max_iter}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def predict(self, X):
        return self._sklearn_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for LinearSVR    Linear Support Vector Regression.',
    'allOf': [{
        'type': 'object',
        'required': ['epsilon', 'tol', 'C', 'loss', 'fit_intercept', 'intercept_scaling', 'dual', 'verbose', 'random_state', 'max_iter'],
        'relevantToOptimizer': ['epsilon', 'tol', 'loss', 'fit_intercept', 'dual', 'max_iter'],
        'additionalProperties': False,
        'properties': {
            'epsilon': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 1.35,
                'distribution': 'loguniform',
                'default': 0.0,
                'description': 'Epsilon parameter in the epsilon-insensitive loss function. Note'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.0001,
                'description': 'Tolerance for stopping criteria.'},
            'C': {
                'type': 'number',
                'default': 1.0,
                'description': 'Penalty parameter C of the error term. The penalty is a squared'},
            'loss': {
                'enum': ['hinge', 'l2', 'squared_epsilon_insensitive', 'squared_hinge', 'epsilon_insensitive'],
                'default': 'epsilon_insensitive',
                'description': 'Specifies the loss function. The epsilon-insensitive loss'},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to calculate the intercept for this model. If set'},
            'intercept_scaling': {
                'type': 'number',
                'default': 1.0,
                'description': 'When self.fit_intercept is True, instance vector x becomes'},
            'dual': {
                'type': 'boolean',
                'default': True,
                'description': 'Select the algorithm to either solve the dual or primal'},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': 'Enable verbose output. Note that this setting takes advantage of a'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The seed of the pseudo random number generator to use when shuffling'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 1000,
                'description': 'The maximum number of iterations to be run.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model according to the given training data.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples in the number of samples and'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target vector relative to X'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'default': None,
            'description': 'Array of weights that are assigned to individual'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the linear model',
    'type': 'object',
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Returns predicted values.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
LinearSVR = lale.operators.make_operator(LinearSVRImpl, _combined_schemas)





from sklearn.naive_bayes import GaussianNB as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class GaussianNBImpl():

    def __init__(self, priors=None, var_smoothing=1e-09):
        self._hyperparams = {
            'priors': priors,
            'var_smoothing': var_smoothing}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def predict(self, X):
        return self._sklearn_model.predict(X)

    def predict_proba(self, X):
        return self._sklearn_model.predict_proba(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for GaussianNB    Gaussian Naive Bayes (GaussianNB)',
    'allOf': [{
        'type': 'object',
        'required': ['priors', 'var_smoothing'],
        'relevantToOptimizer': [],
        'additionalProperties': False,
        'properties': {
            'priors': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }, {
                    'enum': [None]}],
                'default': None,
                'description': 'Prior probabilities of the classes. If specified the priors are not'},
            'var_smoothing': {
                'type': 'number',
                'default': 1e-09,
                'description': 'Portion of the largest variance of all features that is added to'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit Gaussian Naive Bayes according to X, y',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vectors, where n_samples is the number of samples'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values.'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'default': None,
            'description': 'Weights applied to individual samples (1. for unweighted).'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted target values for X',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return probability estimates for the test vector X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Returns the probability of the samples for each class in',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
GaussianNB = lale.operators.make_operator(GaussianNBImpl, _combined_schemas)





from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class QuadraticDiscriminantAnalysisImpl():

    def __init__(self, priors=None, reg_param=0.0, store_covariance=False, tol=0.0001, store_covariances=None):
        self._hyperparams = {
            'priors': priors,
            'reg_param': reg_param,
            'store_covariance': store_covariance,
            'tol': tol,
            'store_covariances': store_covariances}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def predict(self, X):
        return self._sklearn_model.predict(X)

    def predict_proba(self, X):
        return self._sklearn_model.predict_proba(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for QuadraticDiscriminantAnalysis    Quadratic Discriminant Analysis',
    'allOf': [{
        'type': 'object',
        'required': ['priors', 'reg_param', 'store_covariance', 'tol', 'store_covariances'],
        'relevantToOptimizer': ['tol'],
        'additionalProperties': False,
        'properties': {
            'priors': {
                'XXX TODO XXX': 'array, optional, shape = [n_classes]',
                'description': 'Priors on classes',
                'enum': [None],
                'default': None},
            'reg_param': {
                'type': 'number',
                'default': 0.0,
                'description': 'Regularizes the covariance estimate as'},
            'store_covariance': {
                'type': 'boolean',
                'default': False,
                'description': 'If True the covariance matrices are computed and stored in the'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.0001,
                'description': 'Threshold used for rank estimation.'},
            'store_covariances': {
                'anyOf': [{
                    'type': 'boolean'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Deprecated, use `store_covariance`.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model according to the given training data and parameters.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples is the number of samples and'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values (integers)'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return posterior probabilities of classification.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Array of samples/test vectors.'},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Posterior probabilities of classification per class.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
QuadraticDiscriminantAnalysis = lale.operators.make_operator(QuadraticDiscriminantAnalysisImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Dict
import os
import logging

from lale.search.search_space import *
from lale.util.Visitor import Visitor
from lale.search import schema2search_space as opt
from lale.search.HP import search_space_to_hp_expr, search_space_to_hp_str
from lale.search.PGO import PGO

from hyperopt import hp

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from lale.operators import PlannedOperator, OperatorChoice, PlannedIndividualOp, PlannedPipeline

def hyperopt_search_space(op:'PlannedOperator', 
                          schema=None,
                          pgo:Optional[PGO]=None):
    return HPOperatorVisitor.run(op, schema=schema, pgo=pgo)

class HPOperatorVisitor(Visitor):
    pgo:Optional[PGO]
    names:Dict[str,int]

    @classmethod
    def run(cls, 
            op:'PlannedOperator',
            schema=None,
            pgo:Optional[PGO]=None):
        visitor = cls(pgo=pgo)
        accepting_op:Any = op
        return accepting_op.accept(visitor, schema=schema)

    def __init__(self, pgo:Optional[PGO]=None):
        super(HPOperatorVisitor, self).__init__()
        self.pgo = pgo
        self.names = {}

    def get_unique_name(self, name:str)->str:
        if name in self.names:
            counter = self.names[name] + 1
            self.names[name] = counter
            return name + "@" + str(counter)
        else:
            self.names[name] = 0
            return name

    def visitPlannedIndividualOp(self, op:'PlannedIndividualOp', schema=None):
        if schema is None:
            schema = op.hyperparam_schema_with_hyperparams()
        module = op._impl.__module__
        if module is None or module == str.__class__.__module__:
            long_name = op.name()
        else:
            long_name = module + '.' + op.name()
        name = op.name()

        (simp, hp_s) = opt.schemaToSimplifiedAndSearchSpace(long_name, name, schema, pgo=self.pgo)
        if hp_s:
            unique_name = self.get_unique_name(name)

            if os.environ.get("LALE_PRINT_SEARCH_SPACE", "false") == "true":
                print(f"hyperopt search space for {unique_name}: {search_space_to_hp_str(hp_s, unique_name)}")
            return search_space_to_hp_expr(hp_s, unique_name)
        else:
            return None

    visitTrainableIndividualOp = visitPlannedIndividualOp
    visitTrainedIndividualOp = visitPlannedIndividualOp

    def visitPlannedPipeline(self, op:'PlannedPipeline', schema=None):
        search_spaces = [m.accept(self) for m in op.steps()]
        return search_spaces

    visitTrainablePipeline = visitPlannedPipeline
    visitTrainedPipeline = visitPlannedPipeline

    def visitOperatorChoice(self, op:'OperatorChoice', schema=None):
        unique_name:str = self.get_unique_name(op.name())
        search_spaces = hp.choice(unique_name, [{i : m.accept(self)} for i, m in enumerate(op.steps())])
        return search_spaces




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .project import ProjectImpl
import lale.operators

class KeepNumbersImpl(ProjectImpl):
    def __init__(self):
        super(KeepNumbersImpl, self).__init__(columns={'type': 'number'})

_hyperparams_schema = {
  'description': 'Hyperparameter schema for KeepNumbers transformer.',
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints.',
      'type': 'object',
      'additionalProperties': False,
      'relevantToOptimizer': [],
      'properties': {}}]}

_input_fit_schema = {
  'description': 'Input data schema for training KeepNumbers.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {
        'type': 'array',
        'items': {
          'anyOf':[{'type': 'number'}, {'type':'string'}]}}},
    'y': {
      'description': 'Target class labels; the array is over samples.'}}}

_input_predict_schema = {
  'description': 'Input data schema for transformation using KeepNumbers.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {
        'type': 'array',
        'items': {
           'anyOf':[{'type': 'number'}, {'type':'string'}]}}}}}

_output_schema = {
  'description': 'Output data schema for transformed data using KeepNumbers.',
  'type': 'array',
  'items': {
    'type': 'array',
    'items': {
       'anyOf':[{'type': 'number'}]}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.keep_numbers.html',
    'type': 'object',
    'tags': {
        'pre': ['categoricals'],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output': _output_schema }}

if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)

KeepNumbers = lale.operators.make_operator(KeepNumbersImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lale.operators import make_operator

import numpy as np
import pandas as pd
import logging
import lale.helpers as helpers
logging.basicConfig(level=logging.INFO)

class BatchingImpl():
  """Batching trains the given pipeline using batches.
  The batch_size is used across all steps of the pipeline, serializing
  the intermediate outputs if specified.

  Parameters
  ----------
  operator : lale.operators.Pipeline
      A Lale pipeline object that needs to be trained/used for transform or predictions,
      by default None
  batch_size : int, optional
      Batch size to be used for all steps in the pipeline, by default 32
  shuffle : bool, optional
      Shuffle dataset before batching or not, by default True
  num_workers : int, optional
      Number of workers for pytorch dataloader, by default 0
  inmemory : bool, optional
      Whether all the computations are done in memory or intermediate outputs are serialized.

  """
  def __init__(self, operator = None, batch_size = 32, shuffle = True, num_workers = 0, inmemory=False):    
    self.operator = operator
    self.batch_size = batch_size
    self.shuffle = shuffle
    self.num_workers = num_workers
    self.inmemory = inmemory

  def fit(self, X, y = None):
    if self.operator is None:
      raise ValueError("The pipeline object can't be None at the time of fit.")
    data_loader = helpers.create_data_loader(X = X, y = y, batch_size = self.batch_size)
    classes = np.unique(y)
    self.operator = self.operator.fit_with_batches(data_loader, y = classes, serialize = self.inmemory)
    return self

  def transform(self, X, y = None):
    data_loader = helpers.create_data_loader(X = X, y = y, batch_size = self.batch_size)
    transformed_data = self.operator.transform_with_batches(data_loader, serialize = self.inmemory)
    return transformed_data

  def predict(self, X, y = None):
    return self.transform(X, y)

_input_schema_fit = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for fit.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'anyOf': [
        { 'type': 'array',
          'items': {'type': 'number'}},
        { 'type': 'array',
          'items': {
            'type': 'array', 'minItems': 1,
            'items': {'type': 'number'}}}]},
        'y': {
            'type': 'array',
            'items': {'anyOf':[{'type': 'integer'}, {'type':'number'}]}
        }
  }
}

_input_schema_predict = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for predictions.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'anyOf': [
        { 'type': 'array',
          'items': {'type': 'number'}},
        { 'type': 'array',
          'items': {
            'type': 'array', 'minItems': 1,
            'items': {'type': 'number'}}}]},
        'y': {
            'type': 'array',
            'items': {'anyOf':[{'type': 'integer'}, {'type':'number'}]}
        }
  }
}

_output_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for transformed data.'}
  

_hyperparams_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Hyperparameter schema.',
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints.',
      'type': 'object',
      'additionalProperties': False,
      'relevantToOptimizer': ['batch_size'],
      'properties': {
        'operator':{
          'description':'A lale pipeline object to be used inside of batching',
        },
        'batch_size':{
          'description': 'Batch size used for transform.',
          'type': 'integer',
          'default': 64,
          'minimum': 1,
          'distribution': 'uniform',
          'minimumForOptimizer': 32,
          'maximumForOptimizer': 128},
        'shuffle':{
          'type':'boolean',
          'default': False,
          'description': 'Shuffle dataset before batching or not.'
          },
        'num_workers':{
          'type':'integer',
          'default':0,
          'description': 'Number of workers for pytorch dataloader.'
          },
        'inmemory':{
          'type':'boolean',
          'default': False,
          'description': 'Whether all the computations are done in memory or intermediate outputs are serialized.'
          }
          }}]}

_combined_schemas = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Combined schema for expected data and hyperparameters for a transformer for'
                ' a text data transformer based on pre-trained BERT model '
                '(https://github.com/huggingface/pytorch-pretrained-BERT).',
  'type': 'object',
  'tags': {
    'pre': [],
    'op': [],
    'post': []},
  'properties': {
    'input_fit': _input_schema_fit,
    'input_predict': _input_schema_predict,
    'output': _output_schema,
    'hyperparams': _hyperparams_schema } }

if __name__ == "__main__":
    helpers.validate_is_schema(_combined_schemas)

Batching = make_operator(BatchingImpl, _combined_schemas)




from sklearn.linear_model.ransac import RANSACRegressor as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class RANSACRegressorImpl():

    def __init__(self, base_estimator=None, min_samples=None, residual_threshold=None, is_data_valid=None, is_model_valid=None, max_trials=100, max_skips='inf', stop_n_inliers='inf', stop_score='inf', stop_probability=0.99, loss='absolute_loss', random_state=None):
        self._hyperparams = {
            'base_estimator': base_estimator,
            'min_samples': min_samples,
            'residual_threshold': residual_threshold,
            'is_data_valid': is_data_valid,
            'is_model_valid': is_model_valid,
            'max_trials': max_trials,
            'max_skips': max_skips,
            'stop_n_inliers': stop_n_inliers,
            'stop_score': stop_score,
            'stop_probability': stop_probability,
            'loss': loss,
            'random_state': random_state}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def predict(self, X):
        return self._sklearn_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for RANSACRegressor    RANSAC (RANdom SAmple Consensus) algorithm.',
    'allOf': [{
        'type': 'object',
        'required': ['base_estimator', 'min_samples', 'residual_threshold', 'is_data_valid', 'is_model_valid', 'max_trials', 'max_skips', 'stop_n_inliers', 'stop_score', 'stop_probability', 'loss', 'random_state'],
        'relevantToOptimizer': ['min_samples', 'max_trials', 'max_skips', 'stop_n_inliers', 'loss'],
        'additionalProperties': False,
        'properties': {
            'base_estimator': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Base estimator object which implements the following methods:'},
            'min_samples': {
                'XXX TODO XXX': 'int (>= 1) or float ([0, 1]), optional',
                'description': 'Minimum number of samples chosen randomly from original data. Treated',
                'anyOf': [{
                    'type': 'number',
                    'minimumForOptimizer': 0.0,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None},
            'residual_threshold': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Maximum residual for a data sample to be classified as an inlier.'},
            'is_data_valid': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'This function is called with the randomly selected data before the'},
            'is_model_valid': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'This function is called with the estimated model and the randomly'},
            'max_trials': {
                'type': 'integer',
                'minimumForOptimizer': 100,
                'maximumForOptimizer': 101,
                'distribution': 'uniform',
                'default': 100,
                'description': 'Maximum number of iterations for random sample selection.'},
            'max_skips': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.0,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'}],
                'default': inf,
                'description': 'Maximum number of iterations that can be skipped due to finding zero'},
            'stop_n_inliers': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.0,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'}],
                'default': inf,
                'description': 'Stop iteration if at least this number of inliers are found.'},
            'stop_score': {
                'type': 'number',
                'default': inf,
                'description': 'Stop iteration if score is greater equal than this threshold.'},
            'stop_probability': {
                'XXX TODO XXX': 'float in range [0, 1], optional',
                'description': 'RANSAC iteration stops if at least one outlier-free set of the training',
                'type': 'number',
                'default': 0.99},
            'loss': {
                'anyOf': [{
                    'type': 'object',
                    'forOptimizer': False}, {
                    'enum': ['absolute_loss', 'squared_loss']}],
                'default': 'absolute_loss',
                'description': 'String inputs, "absolute_loss" and "squared_loss" are supported which'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The generator used to initialize the centers.  If int, random_state is'},
        }}, {
        'XXX TODO XXX': 'Parameter: base_estimator > only supports regression estimators'}, {
        'XXX TODO XXX': 'Parameter: is_model_valid > only be used if the estimated model is needed for making the rejection decision'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit estimator using RANSAC algorithm.',
    'type': 'object',
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix, shape [n_samples, n_features]'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Training data.'},
        'y': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Target values.'},
        'sample_weight': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Individual weights for each sample'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the estimated model.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Returns predicted values.',
    'anyOf': [{
        'type': 'array',
        'items': {
            'type': 'number'},
    }, {
        'type': 'array',
        'items': {
            'type': 'array',
            'items': {
                'type': 'number'},
        }}],
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
RANSACRegressor = lale.operators.make_operator(RANSACRegressorImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.helpers
import lale.operators
import sklearn.neighbors

class KNeighborsClassifierImpl():
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams

    def fit(self, X, y=None):
        self._sklearn_model = sklearn.neighbors.KNeighborsClassifier(**self._hyperparams)
        self._sklearn_model.fit(X, y)
        return self

    def predict(self, X):
        return self._sklearn_model.predict(X)

    def predict_proba(self, X):
        return self._sklearn_model.predict_proba(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Hyperparameter schema for the KNeighborsClassifier model from scikit-learn.',
    'allOf': [{
        'description':
            'This first sub-object lists all constructor arguments with their '
            'types, one at a time, omitting cross-argument constraints.',
        'type': 'object',
        'additionalProperties': False,
        'required': [
            'n_neighbors', 'weights', 'algorithm', 'leaf_size', 'p', 'metric',
            'metric_params', 'n_jobs'],
        'relevantToOptimizer': [
            'n_neighbors', 'weights', 'algorithm', 'p', 'metric'],
        'properties': {
            'n_neighbors': {
                'description': 'Number of neighbors to use by default for kneighbors queries.',
                'type': 'integer',
                'distribution': 'loguniform',
                'minimum': 1,
                'default': 5,
                'maximumForOptimizer': 100},
            'weights': {
                'description': 'Weight function used in prediction.',
                'enum': ['uniform', 'distance'],
                'default': 'uniform'},
            'algorithm': {
                'description': 'Algorithm used to compute the nearest neighbors.',
                'enum': ['ball_tree', 'kd_tree', 'brute', 'auto'],
                'default': 'auto'},
            'leaf_size': {
                'description': 'Leaf size passed to BallTree or KDTree.',
                'type': 'integer',
                'distribution': 'loguniform',
                'minimum': 1,
                'default': 30},
            'p': {
                'description': 'Power parameter for the Minkowski metric.',
                'type': 'integer',
                'distribution': 'uniform',
                'minimum': 1,
                'default': 2,
                'maximumForOptimizer': 3},
            'metric': {
                'description': 'The distance metric to use for the tree.',
                'enum': ['euclidean', 'manhattan', 'minkowski'],
                'default': 'minkowski'},
            'metric_params': {
                'description': 'Additional keyword arguments for the metric function.',
                'anyOf': [{
                    'enum': [None]}, {
                    'type': 'object',
                    'propertyNames': {
                        'pattern': '[_a-zA-Z][_a-zA-Z0-9]*'},
                }],
                'default': None},
            'n_jobs': {
                'description': 'Number of parallel jobs to run for the neighbor search.',
                'anyOf': [{
                    'description': '1 unless in joblib.parallel_backend context.',
                    'enum': [None]}, {
                    'description': 'Use all processors.',
                    'enum': [(- 1)]}, {
                    'description': 'Number of CPU cores.',
                    'type': 'integer',
                    'minimum': 1}],
                'default': None}}},
    {   'description': 'The leaf size only matters for tree algorithms.',
        'anyOf': [
        {   'type': 'object',
            'properties': {
                'algorithm': {
                    'enum': ['ball_tree', 'kd_tree']},
            }},
        {   'type': 'object',
            'properties': {
                'leaf_size': {
                    'enum': [30]},
            }}]},
    {   'description': 'The power parameter is specific to the minkowski metric.',
        'anyOf': [
        {   'type': 'object',
            'properties': {
                'metric': {
                    'enum': ['minkowski']},
            }},
        {   'type': 'object',
            'properties': {
                'p': {
                    'enum': [2]},
            }}]}]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training the KNeighborsClassifier model from scikit-learn.',
    'type': 'object',
    'required': ['X', 'y'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
        'y': {
            'description': 'Target class labels; the array is over samples.',
            'anyOf': [
            {   'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }},
            {   'type': 'array',
                'items': {
                    'type': 'number'},
            }]}}}

_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for predictions using the KNeighborsClassifier model from scikit-learn.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': { 'type': 'number'}}}}}

_output_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Output data schema for predictions (target class labels) using the KNeighborsClassifier model from scikit-learn.',
    'anyOf': [
    {   'type': 'array',
        'items': {
            'type': 'array',
            'items': {
                'type': 'number'},
        }},
    {   'type': 'array',
        'items': {
            'type': 'number'}}]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html',
    'type': 'object',
    'tags': {
        'pre': ['~categoricals'],
        'op': ['estimator', 'classifier', 'interpretable'],
        'post': ['probabilities']},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output': _output_schema,
    },
}

if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)

KNeighborsClassifier = lale.operators.make_operator(KNeighborsClassifierImpl, _combined_schemas)




from sklearn.cluster.k_means_ import KMeans as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class KMeansImpl():

    def __init__(self, n_clusters=8, init='k-means++', n_init=10, max_iter=300, tol=0.0001, precompute_distances='auto', verbose=0, random_state=None, copy_x=True, n_jobs=None, algorithm='auto'):
        self._hyperparams = {
            'n_clusters': n_clusters,
            'init': init,
            'n_init': n_init,
            'max_iter': max_iter,
            'tol': tol,
            'precompute_distances': precompute_distances,
            'verbose': verbose,
            'random_state': random_state,
            'copy_x': copy_x,
            'n_jobs': n_jobs,
            'algorithm': algorithm}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def transform(self, X):
        return self._sklearn_model.transform(X)

    def predict(self, X):
        return self._sklearn_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for KMeans    K-Means clustering',
    'allOf': [{
        'type': 'object',
        'required': ['n_clusters', 'init', 'n_init', 'max_iter', 'tol', 'precompute_distances', 'verbose', 'random_state', 'copy_x', 'n_jobs', 'algorithm'],
        'relevantToOptimizer': ['n_clusters', 'init', 'n_init', 'max_iter', 'tol', 'precompute_distances', 'copy_x', 'algorithm'],
        'additionalProperties': False,
        'properties': {
            'n_clusters': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 8,
                'distribution': 'uniform',
                'default': 8,
                'description': 'The number of clusters to form as well as the number of'},
            'init': {
                'enum': ['k-means++', 'random', 'ndarray'],
                'default': 'k-means++',
                'description': "Method for initialization, defaults to 'k-means++':"},
            'n_init': {
                'type': 'integer',
                'minimumForOptimizer': 3,
                'maximumForOptimizer': 10,
                'distribution': 'uniform',
                'default': 10,
                'description': 'Number of time the k-means algorithm will be run with different'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 300,
                'description': 'Maximum number of iterations of the k-means algorithm for a'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.0001,
                'description': 'Relative tolerance with regards to inertia to declare convergence'},
            'precompute_distances': {
                'enum': ['auto', True, False],
                'default': 'auto',
                'description': 'Precompute distances (faster but takes more memory).'},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': 'Verbosity mode.'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Determines random number generation for centroid initialization. Use'},
            'copy_x': {
                'type': 'boolean',
                'default': True,
                'description': 'When pre-computing distances it is more numerically accurate to center'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of jobs to use for the computation. This works by computing'},
            'algorithm': {
                'XXX TODO XXX': '"auto", "full" or "elkan", default="auto"',
                'description': 'K-means algorithm to use. The classical EM-style algorithm is "full".',
                'enum': ['auto', 'elkan', 'full'],
                'default': 'auto'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Compute k-means clustering.',
    'type': 'object',
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix, shape=(n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Training instances to cluster. It must be noted that the data'},
        'y': {
            'description': 'not used, present here for API consistency by convention.'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'default': None,
            'description': 'The weights for each observation in X. If None, all observations'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform X to a cluster-distance space.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'New data to transform.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'X transformed in the new space.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict the closest cluster each sample in X belongs to.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'New data to predict.'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'default': None,
            'description': 'The weights for each observation in X. If None, all observations'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Index of the cluster each sample belongs to.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
KMeans = lale.operators.make_operator(KMeansImpl, _combined_schemas)





from sklearn.linear_model.bayes import ARDRegression as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class ARDRegressionImpl():

    def __init__(self, n_iter=300, tol=0.001, alpha_1=1e-06, alpha_2=1e-06, lambda_1=1e-06, lambda_2=1e-06, compute_score=False, threshold_lambda=10000.0, fit_intercept=True, normalize=False, copy_X=True, verbose=False):
        self._hyperparams = {
            'n_iter': n_iter,
            'tol': tol,
            'alpha_1': alpha_1,
            'alpha_2': alpha_2,
            'lambda_1': lambda_1,
            'lambda_2': lambda_2,
            'compute_score': compute_score,
            'threshold_lambda': threshold_lambda,
            'fit_intercept': fit_intercept,
            'normalize': normalize,
            'copy_X': copy_X,
            'verbose': verbose}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def predict(self, X):
        return self._sklearn_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for ARDRegression    Bayesian ARD regression.',
    'allOf': [{
        'type': 'object',
        'required': ['n_iter', 'tol', 'alpha_1', 'alpha_2', 'lambda_1', 'lambda_2', 'compute_score', 'threshold_lambda', 'fit_intercept', 'normalize', 'copy_X', 'verbose'],
        'relevantToOptimizer': ['n_iter', 'tol', 'compute_score', 'fit_intercept', 'normalize', 'copy_X'],
        'additionalProperties': False,
        'properties': {
            'n_iter': {
                'type': 'integer',
                'minimumForOptimizer': 5,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 300,
                'description': 'Maximum number of iterations. Default is 300'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.001,
                'description': 'Stop the algorithm if w has converged. Default is 1.e-3.'},
            'alpha_1': {
                'type': 'number',
                'default': 1e-06,
                'description': 'Hyper-parameter : shape parameter for the Gamma distribution prior'},
            'alpha_2': {
                'type': 'number',
                'default': 1e-06,
                'description': 'Hyper-parameter : inverse scale parameter (rate parameter) for the'},
            'lambda_1': {
                'type': 'number',
                'default': 1e-06,
                'description': 'Hyper-parameter : shape parameter for the Gamma distribution prior'},
            'lambda_2': {
                'type': 'number',
                'default': 1e-06,
                'description': 'Hyper-parameter : inverse scale parameter (rate parameter) for the'},
            'compute_score': {
                'type': 'boolean',
                'default': False,
                'description': 'If True, compute the objective function at each step of the model.'},
            'threshold_lambda': {
                'type': 'number',
                'default': 10000.0,
                'description': 'threshold for removing (pruning) weights with high precision from'},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'whether to calculate the intercept for this model. If set'},
            'normalize': {
                'type': 'boolean',
                'default': False,
                'description': 'This parameter is ignored when ``fit_intercept`` is set to False.'},
            'copy_X': {
                'type': 'boolean',
                'default': True,
                'description': 'If True, X will be copied; else, it may be overwritten.'},
            'verbose': {
                'type': 'boolean',
                'default': False,
                'description': 'Verbose mode when fitting the model.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the ARDRegression model according to the given training data',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples in the number of samples and'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': "Target values (integers). Will be cast to X's dtype if necessary"},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the linear model.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Samples.'},
        'return_std': {
            'anyOf': [{
                'type': 'boolean'}, {
                'enum': [None]}],
            'default': None,
            'description': 'Whether to return the standard deviation of posterior prediction.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the linear model.',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
ARDRegression = lale.operators.make_operator(ARDRegressionImpl, _combined_schemas)

